"""
Tools for visualizing qcio objects in Jupyter Notebooks.

Design Decisions:
    - The view function is the top-level method for viewing all qcio objects. It can 
        accept one or many objects and will determine the appropriate viewer to use.
    - All functions that begin with "generate" will return a string of HTML. If users
        want to use this HTML to create a custom view, they can do so. If they want to
        display the HTML in a Jupyter Notebook, they can call display(HTML(html_string))
        after importing `from IPython.display import HTML, display`.
    - The basic layout for viewing results (all ProgramOutput objects) is a table of
        basic parameters followed by a structure viewer and results table or plot.
        DualProgramInputs add details for the subprogram.
        ----------------------------------------------------------------------------
        | Structure Name | Success | Calculation Type | Program | Model | Keywords |
        ----------------------------------------------------------------------------
        |                                    |                                     |
        |      Structure Viewer (Optional)   |        Results Table or Plot        | 
        |                                    |                                     |
        ----------------------------------------------------------------------------
"""

import base64
import html
import io
from contextlib import contextmanager
from itertools import zip_longest
from typing import Any, Dict, List, Optional, Union

import numpy as np

from qcio import (
    DualProgramInput,
    OptimizationResults,
    ProgramInput,
    ProgramOutput,
    SinglePointResults,
    Structure,
    constants,
)

# View dependencies
try:
    import matplotlib.pyplot as plt
    import py3Dmol as p3d
    from IPython.display import HTML, display
except ImportError as e:
    missing_packages = []
    if "matplotlib" in str(e):
        missing_packages.append("matplotlib")
    if "py3Dmol" in str(e):
        missing_packages.append("py3Dmol")
    if "IPython" in str(e):
        missing_packages.append("IPython")
    raise ImportError(
        f"Missing dependencies: {', '.join(missing_packages)} required for the view "
        "module. Please install them using: pip install qcio[view]"
    )


def generate_structure_viewer_html(
    struct: Union[Structure, List[Structure]],
    *,
    width: int = 600,
    height: int = 450,
    style: Optional[Dict] = None,
    show_indices: bool = False,
    png: bool = False,
    **kwargs,
) -> str:
    """
    Generate the core HTML viewer for a Structure or list of Structures using py3Dmol.

    Args:
        struct: The Structure or list of Structures to visualize. Lists will be
            animated.
        width: The width of the viewer in pixels.
        height: The height of the viewer in pixels.
        style: A dictionary of styles to apply to the viewer.
        show_indices: Whether to show the atom indices in the viewer.
        png: Whether to return a PNG image of the viewer instead of the WebGL HTML.

    Returns:
        str: The HTML string for the viewer.
    """
    viewer = p3d.view(width=width, height=height)
    if isinstance(struct, list):
        combined_xyz = "".join(s.to_xyz() for s in struct)
        viewer.addModelsAsFrames(combined_xyz, "xyz")
        viewer.animate({"loop": "forward"})
    else:
        viewer.addModel(struct.to_xyz(), "xyz")

    struct = struct[-1] if isinstance(struct, list) else struct

    if show_indices:
        for j, (symbol, coord) in enumerate(
            zip(struct.symbols, struct.geometry * constants.BOHR_TO_ANGSTROM)
        ):
            viewer.addLabel(
                str(j),
                {
                    "position": {"x": coord[0], "y": coord[1], "z": coord[2]},
                    "style": {"fontColor": "black", "backgroundColor": "white"},
                },
            )

    viewer.setStyle(style or {"stick": {}, "sphere": {"scale": 0.3}})
    viewer.zoomTo()

    return (
        viewer.write_html()
        if not png
        else NotImplementedError("PNG generation is not yet implemented.")
    )


def generate_structure_html(
    struct: Union[Structure, List[Structure]],
    *,
    title: Optional[str] = None,
    subtitle: Optional[str] = None,
    title_extra: Optional[str] = None,
    subtitle_extra: Optional[str] = None,
    **kwargs,
) -> str:
    """
    Generate an HTML viewer with an optional title and/or subtitle for a Structure or
    list of Structures using py3Dmol.

    Args:
        struct: The Structure or list of Structures to visualize. Lists will be
            animated.
        title: The title to display above the viewer.
        subtitle: The subtitle to display below the viewer.
        title_extra: Extra text to display after the title.
        subtitle_extra: Extra text to display after the subtitle.
        **kwargs: Additional keyword arguments to pass to
            generate_structure_viewer_html.

    Returns:
        str: The HTML string for the viewer with optional titles and subtitles.
    """
    mol_viewer_html = generate_structure_viewer_html(struct, **kwargs)
    struct = struct[0] if isinstance(struct, list) else struct

    title = f"{title or struct.ids.name or struct.ids.smiles or ''}"
    title_extra = f"{title_extra or ''}"
    subtitle = f"{subtitle or ''}"
    subtitle_extra = f"{subtitle_extra or ''}"

    return f""" 
        <div style="text-align: center;">
            <h3>{title} {title_extra}</h3>
            {mol_viewer_html}
            <h4>{subtitle} {subtitle_extra}</h4>
        </div>
    """


def generate_dictionary_string(dictionary: Dict[str, Any]) -> str:
    """
    Generate an HTML string displaying a dictionary without explicit row colors.

    Args:
        dictionary: The dictionary to display.

    Returns:
        str: A string of HTML displaying the dictionary with each key/value pair on a
            new line.
    """
    rows = "".join(
        f"<tr><td class='key'>{key}</td><td class='value'>{value}</td></tr>"
        for key, value in dictionary.items()
    )
    return f"<table class='inner-table'>{rows}</table>"


def generate_files_string(files: Dict[str, Union[str, bytes]]) -> str:
    """
    Generate an HTML string displaying a dictionary of files substituting <str> or
    <bytes> for the actual file data
    """
    # Generate the symbolic dictionary
    viewer_dict = {}
    for key, value in files.items():
        if isinstance(value, bytes):
            viewer_dict[key] = html.escape("<bytes>")
        elif isinstance(value, str):
            viewer_dict[key] = html.escape("<str>")

    return generate_dictionary_string(viewer_dict)


def generate_output_table(*prog_outputs: ProgramOutput) -> str:
    """
    Generate an HTML table displaying the basic parameters for ProgramOutput objects.

    Args:
        *prog_outputs: The ProgramOutput objects to display.

    Returns:
        str: A string of HTML displaying the ProgramOutput objects in a table.
    """
    styles = """
    <style>
        table {
            border-collapse: collapse;
            width: 100%;
        }
        th, td {
            border: 1px solid black;
            padding: 8px;
            text-align: left;
        }
        .inner-table {
            border: none;
            width: 100%;
        }
        .inner-table td {
            border: none;
            padding: 2px 5px;
        }
        .key, .value {
            text-align: left;
            white-space: nowrap;
        }
    </style>
    """

    table_header = """
    <table>
        <tr>
            <th>Structure Name</th>
            <th>Success</th>
            <th>Wall Time</th>
            <th>Calculation Type</th>
            <th>Program</th>
            <th>Model</th>
            <th>Keywords</th>
    """
    if any(po.input_data.files for po in prog_outputs):
        table_header += "<th>Input Files</th>"

    if any(isinstance(po.input_data, DualProgramInput) for po in prog_outputs):
        table_header += """
            <th>Subprogram</th>
            <th>Subprogram Model</th>
            <th>Subprogram Keywords</th>
        """
    table_header += "</tr>"

    table_rows = []
    for po in prog_outputs:
        success_style = (
            'style="color: green; font-weight: bold;"'
            if po.success
            else 'style="color: red; font-weight: bold;"'
        )
        base_row = f"""
        <tr>
            <td>{po.input_data.structure.ids.name}</td>
            <td {success_style}>{po.success}</td>
            <td> {
                _format_time(po.provenance.wall_time) if po.provenance.wall_time 
                 else "No timing data"}</td>
            <td>{po.input_data.calctype.name}</td>
            <td>{f"{po.provenance.program} {po.provenance.program_version or ''}"}</td>
            <td>{po.input_data.model}</td>
            <td>{generate_dictionary_string(po.input_data.keywords)}</td>
        """
        if po.input_data.files:
            base_row += f"<td>{generate_files_string(po.input_data.files)}</td>"

        if isinstance(po.input_data, DualProgramInput):
            base_row += f"""
            <td>{po.input_data.subprogram}</td>
            <td>{po.input_data.subprogram_args.model}</td>
            <td>{generate_dictionary_string(po.input_data.subprogram_args.keywords)}</td>
            """
        base_row += "</tr>"
        table_rows.append(base_row)

    table_footer = "</table>"
    return styles + table_header + "\n".join(table_rows) + table_footer


def generate_optimization_plot(
    prog_output: ProgramOutput, figsize=(6.4, 4.8), grid=True
) -> str:
    """
    Generate a plot of the energy optimization by cycle for a single ProgramOutput.

    Args:
        prog_output: The ProgramOutput to generate the plot for.
        figsize: The size of the figure in inches.
        grid: Whether to display grid lines on the plot.

    Returns:
        str: A string of HTML displaying the plot as a png image encoded in base64.
    """
    energies = prog_output.results.energies * constants.HARTREE_TO_KCAL_PER_MOL
    baseline_energy = energies[0]
    relative_energies = energies - baseline_energy
    last_is_nan = np.isnan(relative_energies[-1])

    if last_is_nan:
        try:
            delta_E = relative_energies[-2]
        except IndexError:
            # If there is only one energy point, set delta_E to nan
            delta_E = np.nan
    else:
        delta_E = relative_energies[-1]

    fig, ax1 = plt.subplots(figsize=figsize)
    color = "tab:blue"
    ax1.set_xlabel("Optimization Cycle")
    ax1.set_ylabel("Relative Energy (kcal/mol)", color=color)
    ax1.plot(relative_energies, label="Energy", marker="o", color="green")
    if last_is_nan:
        ax1.plot(len(relative_energies) - 1, delta_E, marker="x", color="red")
    ax1.tick_params(axis="y", labelcolor=color)
    ax1.text(
        0.95,
        0.85,
        f"ΔE = {delta_E:.2f} kcal/mol",
        verticalalignment="top",
        horizontalalignment="right",
        transform=ax1.transAxes,
        color="black",
        fontsize=14,
    )
    plt.title("Energy Optimization by Cycle", pad=20)
    ax1.legend(loc="upper right")
    fig.tight_layout(rect=[0, 0, 1, 0.95])

    buf = io.BytesIO()
    plt.savefig(buf, format="png", bbox_inches="tight")
    buf.seek(0)
    image_base64 = base64.b64encode(buf.read()).decode("utf-8")
    buf.close()
    plt.close(fig)  # Close the figure to avoid duplicate plots
    return (
        f'<img src="data:image/png;base64,{image_base64}" alt="Energy Optimization by '
        'Cycle" style="width: 100%; max-width: 600px;">'
    )


@contextmanager
def _numpy_print_options(*args, **kwargs):
    """
    Temporarily set numpy print options.

    Args:
        *args: Arguments for numpy print options.
        **kwargs: Keyword arguments for numpy print options.
    """
    original_options = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    try:
        yield
    finally:
        np.set_printoptions(**original_options)


def _not_empty(value) -> bool:
    """
    Check if a value is empty. Accommodates numpy arrays.

    Args:
        value: The value to check.

    Returns:
        bool: True if the value is not empty, False otherwise.
    """
    if isinstance(value, np.ndarray):
        return bool(value.any())
    return bool(value)


def _format_time(seconds_float: float) -> str:
    """Format a time in seconds to a human-readable string."""
    hours = int(seconds_float // 3600)
    minutes = int((seconds_float % 3600) // 60)
    seconds = seconds_float % 60
    if hours > 0:
        formatted_time = f"{hours:02}h:{minutes:02}m:{seconds:05.2f}s"
    elif minutes > 0:
        formatted_time = f"{minutes:02}m:{seconds:05.2f}s"
    else:
        formatted_time = f"{seconds:05.2f}s"

    return formatted_time


def generate_results_table(results: SinglePointResults) -> str:
    """
    Generate an HTML table displaying the results.

    Args:
        results: The Results object to display.

    Returns:
        str: A string of HTML displaying the results in a table.
    """

    rows = ""
    for key, value in results.__dict__.items():
        if _not_empty(value):
            if key != "files":
                with _numpy_print_options(
                    threshold=10,
                    formatter={"float_kind": "{:0.2e}".format},
                ):
                    rows += f"<tr><td>{key}</td><td>{value}</td></tr>"

    # Add the files to the bottom table
    if _not_empty(results.files):
        rows += (
            f"<tr><td>Files</td><td>{generate_files_string(results.files)}</td></tr>"
        )

    return f"""
    <table style="width: 100%; border-collapse: collapse; text-align: left;">
        <thead>
            <tr>
                <th style="border: 1px solid black; padding: 8px;">Result</th>
                <th style="border: 1px solid black; padding: 8px;">Value</th>
            </tr>
        </thead>
        <tbody>
            {rows}
        </tbody>
    </table>
    """


def structures(
    *structures: Structure,
    titles: Optional[List[str]] = None,
    subtitles: Optional[List[str]] = None,
    titles_extra: Optional[List[str]] = None,
    subtitles_extra: Optional[List[str]] = None,
    width: int = 600,
    height: int = 450,
    show_indices: bool = False,
    **kwargs,
) -> None:
    """
    Display one or many Structure objects.

    Args:
        *structures: The Structure objects to display.
        titles: The titles to display above the structures.
        subtitles: The subtitles to display below the structures.
        width: The width of the structure viewer in pixels.
        height: The height of the structure viewer in pixels.
        show_indices: Whether to show the atom indices in the viewer.
        **kwargs: Additional keyword arguments to pass to the viewer functions.

    Returns:
        None. Displays the structures in the Jupyter Notebook.
    """
    titles = titles or []
    subtitles = subtitles or []
    titles_extra = titles_extra or []
    subtitles_extra = subtitles_extra or []

    final_html = []

    for i, (struct, title, subtitle, t_extra, s_extra) in enumerate(
        zip_longest(structures, titles, subtitles, titles_extra, subtitles_extra)
    ):
        if i % 2 == 0:
            if i != 0:
                final_html.append("</div>")
            final_html.append(
                '<div style="display: flex; align-items: center; justify-content: center;">'  # noqa
            )

        final_html.append(
            generate_structure_html(
                struct,
                width=width,
                height=height,
                title=title,
                subtitle=subtitle,
                title_extra=t_extra,
                subtitle_extra=s_extra,
                show_indices=show_indices,
                **kwargs,
            )
        )

    if len(structures) % 2 != 0:
        final_html.append("</div>")

    display(HTML("".join(final_html)))


def program_outputs(
    *prog_outputs: ProgramOutput[
        Union[ProgramInput, DualProgramInput], OptimizationResults
    ],
    width: int = 600,
    height: int = 450,
    animate: bool = True,
    show_indices: bool = False,
    struct_viewer: bool = True,
    **kwargs,
) -> None:
    """
    Display one or many ProgramOutput objects.

    Args:
        *prog_outputs: The ProgramOutput objects to display.
        width: The width of the structure and plot viewer in pixels.
        height: The height of the structure and plot viewer in pixels.
        animate: Whether to animate the structure if it is an optimization.
        show_indices: Whether to show the atom indices in the viewer.
        **kwargs: Additional keyword arguments to pass to the viewer functions.

    Returns:
        None. Displays the ProgramOutput objects in the Jupyter Notebook.
    """
    titles = kwargs.get("titles") or []
    subtitles = kwargs.get("subtitles") or []
    titles_extra = kwargs.get("titles_extra") or []
    subtitles_extra = kwargs.get("subtitles_extra") or []

    final_html = []
    for po, title, subtitle, t_extra, s_extra in zip_longest(
        prog_outputs, titles, subtitles, titles_extra, subtitles_extra
    ):
        final_html.append(generate_output_table(po))

        # Create structure viewer
        if not struct_viewer:
            structure_html = "struct_viewer = False"

        else:
            # Determine the structure to use
            title_extra = t_extra or ""
            if isinstance(po.results, OptimizationResults):
                for_viewer: Union[Structure, List[Structure]]
                if animate:
                    for_viewer = po.results.structures
                else:
                    for_viewer = po.results.final_structure
                    title_extra = "(Final Structure)"
            else:
                for_viewer = po.input_data.structure

            structure_html = generate_structure_html(
                for_viewer,
                title=title,
                title_extra=title_extra,
                subtitle=subtitle,
                subtitle_extra=s_extra,
                show_indices=show_indices,
                **kwargs,
            )

        # Create results table or plot
        if isinstance(po.results, OptimizationResults):
            results_html = generate_optimization_plot(
                po, figsize=(width / 100, height / 100)
            )
        else:
            results_html = generate_results_table(po.results)

        final_html.append(
            f"""
    <div style="text-align: center;">
        <div style="display: flex; align-items: center; justify-content: space-around;">
            <div style="text-align: center; margin-right: 20px; flex: 1;">
                <div style="display: inline-block; text-align: center;">
                    {structure_html}
                </div>
            </div>
            <div style="text-align: center; margin-left: 20px; flex: 1;">
                {results_html}
            </div>
        </div>
    </div>
            """
        )

    display(HTML("".join(final_html)))


def view(
    *objs: Union[ProgramOutput, Structure],
    titles: Optional[List[str]] = None,
    subtitles: Optional[List[str]] = None,
    **kwargs,
) -> None:
    """
    Top level method for viewing all qcio objects.

    Args:
        *objs: The ProgramOutput objects to view. May pass one or more objects.
        titles: The titles to display above the structures.
        subtitles: The subtitles to display below the structures.
        **kwargs: Additional keyword arguments to pass to the viewer functions.

    Returns:
        None. Displays the qcio objects in the Jupyter Notebook.
    """
    titles = titles or []
    subtitles = subtitles or []

    if len(titles) > len(objs):
        raise ValueError("Number of titles exceeds number of objects.")
    if len(subtitles) > len(objs):
        raise ValueError("Number of subtitles exceeds number of objects.")

    if all(isinstance(obj, Structure) for obj in objs):
        structures(*objs, titles=titles, subtitles=subtitles, **kwargs)  # type: ignore
        return

    for obj, title, subtitle in zip_longest(objs, titles, subtitles):
        _titles = [title] if title else None
        _subtitles = [subtitle] if subtitle else None

        if isinstance(obj, Structure):
            structures(obj, titles=_titles, subtitles=_subtitles, **kwargs)

        elif isinstance(obj, ProgramOutput):
            program_outputs(obj, titles=_titles, subtitles=_subtitles, **kwargs)

        else:
            raise NotImplementedError(f"Viewing of {type(obj)} is not implemented.")



"""Input models for quantum chemistry calculations."""

import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, field_serializer
from typing_extensions import Self, TypeVar

from .base_models import CalcType, Files, Model
from .structure import Structure

__all__ = [
    "FileInput",
    "ProgramInput",
    "DualProgramInput",
    "ProgramArgs",
    "Inputs",
    "InputType",
    "StructuredInputs",
    "ProgramArgsSub",
]


class FileInput(Files):
    """File and command line argument inputs for a calculation.

    Attributes:
        files: A dict mapping filename to str or bytes data.
        cmdline_args: A list of command line arguments to be passed to the program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    cmdline_args: List[str] = []

    @classmethod
    def from_directory(cls, directory: Union[Path, str], **kwargs) -> Self:
        """Collect all files from directory and instantiate an object."""
        obj = cls(**kwargs)
        directory = Path(directory)
        obj.add_files(directory)
        return obj


class _KeywordsMixin(BaseModel):
    keywords: Dict[str, Any] = {}


class _StructureKeywordsMixin(_KeywordsMixin):
    """
    Attributes:
        structure: The structure to be used in the calculation or SMILES string.
    """

    structure: Structure

    def __init__(self, **data: Any):
        """Backwards compatibility for 'molecule' attribute.

        TODO: Remove in future versions.
        """
        if "molecule" in data:
            warnings.warn(
                "Use of 'molecule' attribute is deprecated. Use 'structure' instead.",
                FutureWarning,
                stacklevel=2,
            )
            data["structure"] = data.pop("molecule")
        super().__init__(**data)

    @property
    def molecule(self) -> Structure:
        """Backwards compatibility for 'molecule' attribute."""
        warnings.warn(
            "Use of 'molecule' attribute is deprecated. Use 'structure' instead.",
            FutureWarning,
            stacklevel=2,
        )
        return self.structure


class ProgramArgs(FileInput, _KeywordsMixin):
    """Generic arguments for a program without a calctype or structure specification.

    This class is needed for multi-step calculations where the calctype and structure
    are specified only once for the entire calculation, e.g., multistep_opt in BigChem.

    Attributes:
        model: The model for the quantum chemistry calculation.
        keywords: A dict of keywords to be passed to the program excluding model and
            calctype. Defaults to an empty dict.
        files: Files to be passed to the QC program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    model: Model


class ProgramArgsSub(FileInput, _KeywordsMixin):
    """Generic arguments for a program that also calls a subprogram.

    This class is needed for multi-step calculations where the calctype and structure
    are specified only once for the entire calculation, e.g., multistep_opt in BigChem.

    Attributes:
        keywords: A dict of keywords to be passed to the program excluding model and
            calctype. Defaults to an empty dict.
        files: Files to be passed to the QC program.
        subprogram: The name of the subprogram to use.
        subprogram_args: The ProgramArgs for the subprogram.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    model: Optional[Model] = None
    subprogram: str
    subprogram_args: ProgramArgs


class ProgramInput(ProgramArgs, _StructureKeywordsMixin):
    """Input for a single quantum chemistry program.

    Attributes:
        calctype: The type of calculation to perform.
        model: The model for the quantum chemistry calculation.
        keywords: A dict of keywords to be passed to the program excluding model and
            calctype. Defaults to an empty dict.
        structure: The structure to be used in the calculation.
        files: Files to be passed to the QC program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    calctype: CalcType

    @field_serializer("calctype")
    def serialize_calctype(self, calctype: CalcType, _info) -> str:
        """Serialize CalcType to string"""
        return calctype.value


class DualProgramInput(ProgramArgsSub, ProgramInput):
    """Input for a two program calculation.

    Attributes:
        calctype: The type of calculation to be performed.
        structure: The structure to be used in the calculation.
        keywords: Dict of keywords to be passed to the program. Defaults to empty dict.
        files: A dict mapping filename to str or bytes data.
        subprogram: The name of the subprogram to use.
        subprogram_args: The program input for the subprogram.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    pass


Inputs = Union[FileInput, ProgramInput, DualProgramInput]
InputType = TypeVar("InputType", bound=Inputs)
StructuredInputs = Union[ProgramInput, DualProgramInput]



"""Physical constants and periodic table from NIST CODATA 2022. All values should have a
reference to their source.

References:
- https://physics.nist.gov/cuu/Constants/Table/allascii.txt
- PubChem periodic table JSON: https://pubchem.ncbi.nlm.nih.gov/periodic-table/
"""

import csv
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel

# https://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0
BOHR_TO_ANGSTROM = 0.529177210544
ANGSTROM_TO_BOHR = 1 / BOHR_TO_ANGSTROM
# https://www.physics.nist.gov/cgi-bin/cuu/Value?hr
HARTREE_TO_JOULE = 4.3597447222060e-18
# https://physics.nist.gov/cgi-bin/cuu/Value?na
AVOGADRO_NUMBER = 6.02214076e23
# https://www.nist.gov/pml/special-publication-811/nist-guide-si-appendix-b-conversion-factors/nist-guide-si-appendix-b8
# Using kcal(th) as Therochemical calorie is the unit used in quantum chemistry.
KCAL_TO_JOULE = 4.184e3
HARTREE_TO_KCAL_PER_MOL = HARTREE_TO_JOULE / KCAL_TO_JOULE * AVOGADRO_NUMBER

_DATA_DIR = Path(__file__).parent / "constants_data"


class Atom(BaseModel):
    """Atom data model."""

    symbol: str
    number: int
    name: str
    mass: float
    group: Optional[int]
    period: int
    block: str
    electron_config: str


class PeriodicTable:
    """Periodic table data model. Data source and URL should be set by class methods."""

    # Should be set by @classmethod

    data_source: str = ""
    data_url: str = ""

    @classmethod
    def from_pubchem(cls):
        """Create a periodic table from PubChem data."""
        instance = cls()
        instance.data_source = "PubChem"
        instance.data_url = "https://pubchem.ncbi.nlm.nih.gov/periodic-table/"

        group_and_period_data = {}

        with open(_DATA_DIR / "group_period.csv", mode="r") as file:
            reader = csv.DictReader(file)
            for row in reader:
                group_and_period_data[row["Symbol"]] = {
                    "group": int(row["Group"]) if row["Group"] else None,
                    "period": int(row["Period"]),
                }

        with open(_DATA_DIR / "pubchem.csv", mode="r") as file:
            reader = csv.DictReader(file)
            for row in reader:
                symbol = row["Symbol"]
                atom = Atom(
                    symbol=symbol,
                    number=int(row["AtomicNumber"]),
                    name=row["Name"],
                    mass=float(row["AtomicMass"]),
                    group=group_and_period_data[symbol]["group"],
                    period=group_and_period_data[symbol]["period"],
                    block=row["GroupBlock"],
                    electron_config=row["ElectronConfiguration"],
                )
                setattr(instance, symbol, atom)
        return instance

    def group(self, group_number: int) -> List[Atom]:
        """Return all atoms in a group."""
        assert 1 <= group_number <= 18, "Group number must be between 1 and 18."
        return [
            atom
            for atom in self.__dict__.values()
            if isinstance(atom, Atom) and atom.group == group_number
        ]

    def period(self, period_number: int) -> List[Atom]:
        """Return all atoms in a period."""
        assert 1 <= period_number <= 7, "Period number must be between 1 and 7."
        return [
            atom
            for atom in self.__dict__.values()
            if isinstance(atom, Atom) and atom.period == period_number
        ]

    def number(self, number: int) -> Atom:
        """Return an atom by atomic number."""
        for atom in self.__dict__.values():
            if isinstance(atom, Atom) and atom.number == number:
                return atom
        raise ValueError(f"No atom with atomic number {number}.")


periodic_table = PeriodicTable.from_pubchem()

# https://depts.washington.edu/eooptic/linkfiles/dielectric_chart%5B1%5D.pdf
_SOLVENTS_DIELECTRIC = {
    "acetic acid": 6.15,
    "acetone": 20.7,
    "acetonitrile": 37.5,
    "anisole": 4.33,
    "benzene": 2.27,
    "bromobenzene": 5.17,
    "carbon disulfide": 2.6,
    "carbon tetrachloride": 2.24,
    "chlorobenzene": 5.62,
    "chloroform": 4.81,
    "cyclohexane": 2.02,
    "dibutyl ether": 3.1,
    "o -dichlorobenzene": 9.93,
    "1,2-dichloroethane": 10.36,
    "dichloromethane": 8.93,
    "diethylamine": 3.6,
    "diethyl ether": 4.33,
    "1,2-dimethoxyethane": 7.2,
    "n,n -dimethylacetamide": 37.8,
    "n,n -dimethylformamide": 36.7,
    "dimethyl sulfoxide": 46.7,
    "1,4-dioxane": 2.25,
    "ethanol": 24.5,
    "ethyl acetate": 6.02,
    "ethyl benzoate": 6.02,
    "formamide": 111,
    "hexamethylphosphoramide": 30,
    "isopropyl alcohol": 17.9,
    "methanol": 32.7,
    "2-methyl-2-propanol": 10.9,
    "nitrobenzene": 34.82,
    "nitromethane": 35.87,
    "pyridine": 12.4,
    "tetrahydrofuran": 7.58,
    "toluene": 2.38,
    "trichloroethylene": 3.4,
    "triethylamine": 2.42,
    "trifluoroacetic acid": 8.55,
    "2,2,2-trifluoroethanol": 8.55,
    "water": 80.1,
    "o -xylene": 2.57,
}



import warnings
from collections import Counter
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import numpy as np
from pydantic import field_serializer, model_validator
from typing_extensions import Self

from qcio.constants import BOHR_TO_ANGSTROM
from qcio.constants import periodic_table as pt
from qcio.helper_types import SerializableNDArray

from .base_models import QCIOModelBase
from .utils import renamed_class, smiles_to_structure, structure_to_smiles

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs

__all__ = ["Structure", "Identifiers", "Molecule"]


class Identifiers(QCIOModelBase):
    """Structure identifiers.

    Attributes:
        name: A human-readable, common name for the structure.
        name_IUPAC: The IUPAC name of the structure.
        smiles: The SMILES representation of the structure.
        canonical_smiles: The canonical SMILES representation of the structure.
        canonical_smiles_program: The program used to generate the canonical SMILES.
        canonical_explicit_hydrogen_smiles: The canonical explicit hydrogen SMILES
            representation of the structure.
        canonical_isomeric_smiles: The canonical isomeric SMILES representation of the
            structure.
        canonical_isomeric_explicit_hydrogen_smiles: The canonical isomeric explicit
            hydrogen SMILES representation of the structure.
        canonical_isomeric_explicit_hydrogen_mapped_smiles: The canonical isomeric
            explicit hydrogen mapped SMILES representation of the structure.
        inchi: The InChI representation of the structure.
        inchikey: The InChIKey representation of the structure.
        pubchem_cid: The PubChem Compound ID of the structure.
        pubchem_sid: The PubChem Substance ID of the structure.
        pubchem_conformerid: The PubChem Conformer ID of the structure.
    """

    name: Optional[str] = None
    name_IUPAC: Optional[str] = None
    smiles: Optional[str] = None
    canonical_smiles: Optional[str] = None
    canonical_smiles_program: Optional[str] = None
    canonical_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_mapped_smiles: Optional[str] = None
    inchi: Optional[str] = None
    inchikey: Optional[str] = None
    pubchem_cid: Optional[str] = None
    pubchem_sid: Optional[str] = None
    pubchem_conformerid: Optional[str] = None


# class Bond(QCIOModelBase):
#     """A bond object.

#     Attributes:
#         indices: The indices of the atoms in the bond.
#         order: The order of the bond.
#     """

#     indices: Tuple[int, int]
#     order: Optional[float]


class Structure(QCIOModelBase):
    """A Structure object with atoms and their corresponding cartesian coordinates;
        also charge, and multiplicity for the entire structure.

    Attributes:
        symbols: The atomic symbols of the structure.
        geometry: The geometry of the structure in Cartesian coordinates. Units are (AU)
            Bohr.
        identifiers: Identifiers for the structure such as name, smiles, etc.
        charge: The molecular charge.
        connectivity: Explicit description of the bonds between atoms.
        multiplicity: The molecular multiplicity.
        connectivity: Explicit description of the bonds between atoms. Each tuple
            contains the indices of the atoms in the bond and the order of the bond.
            E.g., [(0, 1, 1.0), (1, 2, 2.0)] indicates a single bond between atoms 0
            and 1 and a double bond between atoms 1 and 2.
        masses: Explicit masses for the atoms. If not set, default isotopic masses are
            used.
    """

    symbols: List[str]
    geometry: SerializableNDArray  # Coerced to 2D array
    charge: int = 0
    multiplicity: int = 1
    identifiers: Identifiers = Identifiers()
    connectivity: List[Tuple[int, int, float]] = []

    def __init__(self, **data: Any):
        """Backwards compatibility for 'ids' attribute."""
        if identifiers := data.pop("ids", None):
            warnings.warn(
                "Passing 'ids' is deprecated and will be removed in a future "
                "release. Please use 'identifiers' instead. Once instantiated, "
                "you can use structure.ids to access the identifiers as a shortcut.",
                category=FutureWarning,
                stacklevel=2,
            )
            data["identifiers"] = identifiers
        super().__init__(**data)

    @property
    def ids(self) -> Identifiers:
        """Shortcut to access the identifiers."""
        return self.identifiers

    @classmethod
    def from_smiles(
        cls,
        smiles: str,
        *,
        program: str = "rdkit",
        force_field: str = "MMFF94s",
        multiplicity: int = 1,
    ) -> Self:
        """Convert a SMILES string to a 3D Structure.

        Args:
            smiles: The SMILES string.
            program: The program to use for the conversion. Defaults to "rdkit".
            force_field: The force field to use. E.g., UFF, MMFF94, MMFF94s, etc.
            multiplicity: The multiplicity of the structure. Defaults to 1.

        Returns:
            A 3D Structure object with identifiers for SMILES and canonical SMILES.
        """
        dict_repr = smiles_to_structure(smiles, program, force_field)
        dict_repr["multiplicity"] = multiplicity
        return cls(**dict_repr)

    def to_smiles(self, program: str = "rdkit", hydrogens: bool = False) -> str:
        """Generate the canonical SMILES representation of the structure.

        Args:
            program: The program to use for the conversion. Defaults to "rdkit".
            hydrogens: Whether to include hydrogens in the SMILES string. Defaults to
                False.
        """
        return structure_to_smiles(self, program=program, hydrogens=hydrogens)

    def __repr_args__(self) -> "ReprArgs":
        """A helper for __repr__ that returns a list of tuples of the form
        (name, value).
        """
        return [  # pragma: no cover
            ("formula", self.formula),
        ]

    def add_smiles(
        self,
        *,
        smiles: Optional[str] = None,
        canonical_smiles: Optional[str] = None,
        canonical_smiles_program: Optional[str] = None,
        canonical_explicit_hydrogen_smiles: Optional[str] = None,
        program: str = "rdkit",
        hydrogens: bool = False,
    ) -> None:
        """Add SMILES data to the identifiers.

        If no SMILES data is provided, the SMILES will be generated from the structure
        using program.

        Args:
            smiles: The SMILES representation of the structure.
            canonical_smiles: The canonical SMILES representation of the structure.
            canonical_explicit_hydrogen_smiles: The canonical explicit hydrogen SMILES.
            canonical_smiles_program: The program used to generate the canonical SMILES.
            program: The program to use for the conversion. Defaults to "rdkit".
            hydrogens: Whether to include hydrogens in the SMILES string. Defaults to
                False.
        """
        # If no SMILES data is provided, generate it from the structure
        if not smiles and not canonical_smiles:
            smiles = self.to_smiles(program=program, hydrogens=hydrogens)
            canonical_smiles_program = program

            if hydrogens:
                canonical_explicit_hydrogen_smiles = smiles
            else:
                canonical_smiles = smiles

        # Create a new Identifiers instance with the updated values
        new_identifiers = self.identifiers.model_copy(
            update={
                "smiles": smiles,
                "canonical_smiles": canonical_smiles,
                "canonical_smiles_program": canonical_smiles_program,
                "canonical_explicit_hydrogen_smiles": canonical_explicit_hydrogen_smiles,  # noqa: E501
            }
        )

        # Replace the existing identifiers with the new instance
        object.__setattr__(self, "identifiers", new_identifiers)

    @model_validator(mode="before")
    def validate_symbols_and_geometry(cls, values):
        """Ensure symbols are valid atomic symbols and geometry is correct."""
        symbols = [symbol.capitalize() for symbol in values.get("symbols", [])]
        for symbol in symbols:
            if not hasattr(pt, symbol):
                raise ValueError(f"Invalid atomic symbol: '{symbol}'")
        values["symbols"] = symbols

        geometry = values.get("geometry")
        if geometry is not None:
            n_atoms = len(values["symbols"])
            values["geometry"] = np.array(geometry).reshape(n_atoms, 3)

        return values

    @field_serializer("connectivity")
    def serialize_connectivity(self, connectivity, _info) -> List[List[float]]:
        """Serialize connectivity to a list of tuples.

        Cannot have homogeneous data types in .toml files so must cast all values to
        floats.
        """
        return [[float(val) for val in bond] for bond in connectivity]

    @property
    def geometry_angstrom(self) -> np.ndarray:
        """Return the geometry of the structure in Angstrom."""
        return self.geometry * BOHR_TO_ANGSTROM

    @property
    def atomic_numbers(self) -> List[int]:
        """Return the atomic numbers of the atoms in the structure."""
        return [getattr(pt, symbol).number for symbol in self.symbols]

    @property
    def formula(self) -> str:
        """Return the molecular formula of the structure using the Hill System.
        # noqa: E501
        https://chemistry.stackexchange.com/questions/1239/order-of-elements-in-a-formula
        """
        counter_elements = Counter(self.symbols)

        carbon_count = counter_elements.pop("C", 0)
        hydrogen_count = counter_elements.pop("H", 0)

        # Sort remaining elements alphabetically
        sorted_elements = sorted(counter_elements.items())

        if hydrogen_count > 0:
            sorted_elements = [("H", hydrogen_count)] + sorted_elements
        if carbon_count > 0:
            sorted_elements = [("C", carbon_count)] + sorted_elements

        return "".join(
            f"{element}{count if count > 1 else ''}"
            for element, count in sorted_elements
        )

    def model_dump(self, **kwargs) -> Dict[str, Any]:
        """Handle tuple in connectivity"""
        as_dict = super().model_dump(**kwargs)
        # Connectivity may be empty and super().dict() will remove empty values
        if (connectivity := as_dict.get("connectivity")) is not None:
            # Must cast all values to floats as toml cannot handle mixed types
            as_dict["connectivity"] = [
                [float(val) for val in bond] for bond in connectivity
            ]
        return as_dict

    @classmethod
    def open(
        cls,
        filepath: Union[Path, str],
        charge: Optional[int] = None,
        multiplicity: Optional[int] = None,
    ) -> Self:
        """Open a structure from a file.

        Args:
            filepath: The path to the file to open. Maybe a path to a JSON file or an
                XYZ file.
        """
        filepath = Path(filepath)

        if (charge or multiplicity) and filepath.suffix != ".xyz":
            raise ValueError(
                "Charge and multiplicity can only be set when opening an XYZ file."
            )

        if filepath.suffix == ".xyz":
            xyz_str = filepath.read_text()
            return cls.from_xyz(xyz_str, charge=charge, multiplicity=multiplicity)

        return super().open(filepath)

    def save(
        self,
        filepath: Union[Path, str],
        exclude_none=True,
        indent: int = 4,
        **kwargs,
    ) -> None:
        """Save a Structure to a file.

        Args:
            filepath: The path to save the structure to.
            exclude_none: If True, attributes with a value of None will not be written
                to the file.
            **kwargs: Additional keyword arguments to pass to the json serializer.

        Notes:
            If the filepath has a .xyz extension, the structure will be saved to an XYZ
            file.
        """
        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            filepath.write_text(self.to_xyz(**kwargs))
            return
        super().save(filepath, exclude_none, indent, **kwargs)

    def to_xyz(self, precision: int = 17) -> str:
        """Return an xyz string representation of the structure.

        Args:
            precision: The number of decimal places to include in the xyz file. Default
                17 which captures all precision of float64.
        Notes:
            Will add qcio data to the comments line with a qcio_key=value format.
        """

        qcio_data = {  # These get added to comments line (line 2) in xyz file
            "qcio_charge": self.charge,
            "qcio_multiplicity": self.multiplicity,
        }
        assert isinstance(self.geometry, np.ndarray)  # For mypy
        geometry_angstrom = self.geometry * BOHR_TO_ANGSTROM

        xyz_lines = []
        xyz_lines.append(f"{len(self.symbols)}")
        xyz_lines.append(f"{' '.join([f'{k}={v}' for k, v in qcio_data.items()])}")

        # Create a format string using the precision parameter
        format_str = f"{{:2s}} {{: >18.{precision}f}} {{: >18.{precision}f}} {{: >18.{precision}f}}"  # noqa: E501

        for symbol, (x, y, z) in zip(self.symbols, geometry_angstrom):
            xyz_lines.append(format_str.format(symbol, x, y, z))
        xyz_lines.append("")  # Append newline to end of file
        return "\n".join(xyz_lines)

    @classmethod
    def from_xyz(
        cls,
        xyz_str: str,
        *,
        charge: Optional[int] = None,
        multiplicity: Optional[int] = None,
    ) -> Self:
        """Create a Structure from an XYZ file or string.

        Args:
            xyx_str: The XYZ string.
            charge: The molecular charge of the structure. If not provided, will read
                from the XYZ string if set or default to 0.
            multiplicity: The molecular multiplicity of the structure. If not provided,
                will read from the XYZ string if set or default to 1.
        Notes:
            Will read qcio data from the comments line with a qcio_key=value format.
        """

        lines = xyz_str.split("\n")

        num_atoms = int(lines[0])
        qcio_kwargs: Dict[str, Any] = {
            item.split("=")[0].replace("qcio_", ""): item.split("=")[1]
            for item in lines[1].strip().split()
            if item.startswith("qcio_")
        }
        if charge is not None and "charge" in qcio_kwargs:
            raise ValueError("Charge cannot be set in the file and as an argument.")
        if multiplicity is not None and "multiplicity" in qcio_kwargs:
            raise ValueError(
                "Multiplicity cannot be set in the file and as an argument."
            )

        # Set charge and multiplicity if provided
        if charge is not None:
            qcio_kwargs["charge"] = charge
        if multiplicity is not None:
            qcio_kwargs["multiplicity"] = multiplicity

        symbols = []
        geometry = []
        for line in lines[2 : 2 + num_atoms]:
            split_line = line.split()
            symbols.append(split_line[0])
            geometry.append([float(val) / BOHR_TO_ANGSTROM for val in split_line[1:]])

        return cls(symbols=symbols, geometry=geometry, **qcio_kwargs)  # type: ignore


@renamed_class(Structure)
class Molecule(Structure):
    pass
